// Self closing tags
const selfClosingTags = new Set(["link", "meta", "input", "br", "img", "hr"]);

// Set a property
const setProperty = (el, name, value = null) => {
    // Check for style property and string value --> set as css string
    if (name === "style") {
        if (value && typeof value === "object") {
            el.style.cssText = null; // Clear styles
            Object.keys(value).forEach(k => el.style[k] = value[k]);
        }
        else {
            el.style.cssText = value;
        }
    }
    // Check for event listener property
    else if (name.startsWith("on")) {
        el[key.toLowerCase()] = value;
    }
    // Default: set the property value
    else if (name !== "html") {
        // el.setAttribute(key, value)
        el[name] = value;
    }
};

// Check if there are differences between two nodes
const nodesDiffs = (node1, node2) => {
    // Check if nodes have the same type
    if (typeof node1 !== typeof node2) { 
        return true; 
    }
    // Check if nodes are strings
    else if (typeof node1 === "string" && node1 !== node2) { 
        return true; 
    }
    // Check if nodes has not the same tag
    else if (node1.type !== node2.type) { 
        return true; 
    }
    // Default, nodes are the same
    return false;
};

// @public main vdom element creator
const kofi = (type, props, ...children) => {
    if (type && typeof type === "string") {
        return {
            "type": type.toLowerCase().trim(), 
            "props": props || {}, 
            "children": (children || []).flat().filter(c => !!c),
        };
    }
    // Check for function type
    else if (typeof type === "function") {
        return type(props, children);
    }
    // Other type --> throw error
    throw new Error("Invalid element type provided");
};

// Render an element
kofi.render = (parent = null, el) => {
    let node = null;
    // Check for text node
    if (typeof el === "string") {
        node = document.createTextNode(el);
    }
    else {
        // Create the new DOM element and assign the element properties
        node = document.createElement(el.type);
        Object.keys(el.props || {}).forEach(name => {
            name !== "html" && setProperty(node, name, el.props[name]);
        });
        // Check if html property has been provided
        if (typeof el.props?.html === "string") {
            node.innerHTML = el.props.html; //Inject html
        }
        // If no html property has been provided
        else {
            (el.children || []).forEach(child => kofi.render(node, child));
        }
    }
    // Mount the new node
    if (parent) {
        parent.appendChild(node);
    }
    return node;
};

// Convert a VDOM element to string
kofi.stringify = (el, delimiter = "") => {
    if (typeof el === "string") {
        return el;
    }
    // Build attributes of this element
    const attrs = Object.keys(el.props || {})
        .map(name => {
            const value = el.props[name]; //Get attribute value
            // Check for boolean value
            if (value === true) {
                return name; //Return only the property name
            }
            // Check for className attribute
            else if (name === "className" && typeof value === "string") {
                return `class="${value}"`;
            }
            // Check for style and object value
            else if (name === "style" && typeof value === "object") {
                // Convert all style values to string to snake-case format
                const styleValues = Object.keys(value).map(key => {
                    return `${key.split(/(?=[A-Z])/).join("-").toLowerCase()}:${value[key]}`;
                });
                return `style="${styleValues.join(";")}"`;
            }
            // Check for string value --> set the value
            else if (typeof value === "string" && name !== "html") {
                return `${name}="${value}"`;
            }
            return "";
        })
        .filter(v => !!v);
    // Check if this tag is in the non closing tags list
    if (selfClosingTags.has(el.type)) {
        return `<${el.type} ${attrs.join(" ")} />`;
    }
    // Build the content of the element
    const content = el.props?.html ? [el.props.html] : (el.children || []).map(child => {
        return stringify(child, delimiter);
    });
    // Return the element with the content
    return `<${el.type} ${attrs.join(" ")}>${content.join(delimiter || "")}</${el.type}>`;
};

// Update an element
kofi.update = (parent, newNode, oldNode, index) => {
    index = index && typeof index === "number" ? index : 0;
    const child = parent.childNodes[index];
    // Check for no old node --> mount this new element
    if (!oldNode) { 
        return kofi.render(parent, newNode); 
    }
    // If there is not new element --> remove the old element
    else if (!newNode) { 
        return parent.removeChild(child); 
    }
    // If nodes has changed
    else if (nodesDiffs(newNode, oldNode)) {
        return parent.replaceChild(kofi.render(null, newNode), child);
    }
    // Change the properties only if element is not an string
    else if (newNode && typeof newNode !== "string") {
        // Get the full properties values and update the element attributes
        const props = Object.assign({}, newNode.props, oldNode.props);
        Object.keys(props).forEach(name => {
            const newValue = newNode.props[name];
            const oldValue = oldNode.props[name];
            if (name !== "html" && newValue !== oldValue) {
                setProperty(child, name, newValue);
            }
        });
        // Update the children for all element
        const maxLength = Math.max(newNode.children.length, oldNode.children.length);
        for (let i = 0; i < maxLength; i++) {
            kofi.update(child, newNode.children[i], oldNode.children[i], i);
        }
    }
};

// @description
kofi.mount = (parent, fn) => {
    parent.replaceChildren(); // Clear parent element
    let oldVdom = null; // Previous vdom
    return props => {
        const newVdom = fn(props);
        oldVdom ? kofi.update(parent, newVdom, oldVdom) : kofi.render(parent, newVdom);
        oldVdom = newVdom; // Update current vdom
    };
};

// Execute the specified function when the DOM is ready
kofi.ready = fn => {
    // Resolve now if DOM has already loaded
    if (document.readyState !== "loading") {
        return fn();
    }
    // If not, wait for DOMContentLoaded event
    document.addEventListener("DOMContentLoaded", () => {
        return fn();
    });
};

// Join class names
kofi.classNames = (...args) => {
    const names = (args || []).map(item => {
        if (typeof item === "string") {
            return item.split(" ").filter(value => !!value);
        }
        else if (Array.isArray(item)) {
            return item.filter(value => typeof value === "string" && !!value); 
        }
        else if (typeof item === "object") {
            return Object.keys(item || {}).filter(key => !!item[key]);
        }
        // Over value --> return an empty array
        return [];
    });
    return names.flat().join(" ");
};

export default kofi;
